#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright @2024 AI. Inspur Inc.
#
# @author: sunxian <sunxian@inspur.com>
# @date: 2024/07/21
#

import json
import math
import tempfile
from contextlib import nullcontext

import torch
from colossalai.lazy import LazyInitContext
from colossalai.moe.utils import skip_init
from colossalai.utils import get_current_device

from ihp.optim import get_lr_scheduler
from ihp.optim import get_optimizer
from ihp.util import importer
from ihp.util.booster import get_booster
from ihp.util.io import save_json
from ihp.util.logger import logger
from ihp.util.metric import format_numel
from ihp.util.metric import get_model_numel


def create_and_load_model(
    storage, coordinator, args, config, with_optimizer, with_scheduler, extra_config=None, extra_info=""
):
    rank = coordinator.rank
    logger.info(f"rank-{rank} -> init {extra_info} model")

    default_dtype = torch.float16 if args.mixed_precision == "fp16" else torch.bfloat16
    torch.set_default_dtype(default_dtype)

    if args.use_lazy_ctx:
        init_ctx = LazyInitContext(default_device=get_current_device())
    elif args.use_skip_ctx:
        init_ctx = skip_init()
    else:
        init_ctx = nullcontext()

    if coordinator.is_master():
        logger.info(f"init ctx: {init_ctx}")

    with init_ctx:
        logger.info(f"rank-{rank} -> [start]init config({args.config_clz}) from {config}")
        config_clz = importer.from_name_to_clz(args.config_clz)
        if isinstance(config, dict):
            with tempfile.NamedTemporaryFile() as tmp:
                save_json(config, tmp.name)
                model_config = config_clz.from_pretrained(tmp.name, trust_remote_code=True)
        else:
            model_config = config_clz.from_pretrained(config, trust_remote_code=True)

        if extra_config is None:
            extra_config = {}

        if hasattr(model_config, "_flash_attn_2_enabled"):
            extra_config["_flash_attn_2_enabled"] = args.use_flash_attn
        if hasattr(model_config, "_attn_implementation"):
            extra_config["_attn_implementation"] = "flash_attention_2" if args.use_flash_attn else "eager"
        if hasattr(model_config, "use_cache"):
            extra_config["use_cache"] = not args.use_flash_attn
        if hasattr(model_config, "output_router_logits"):
            extra_config["output_router_logits"] = True
        if hasattr(model_config, "router_aux_loss_coef") and args.router_aux_loss_coef > 0.0:
            extra_config["router_aux_loss_coef"] = args.router_aux_loss_coef
        if hasattr(model_config, "initializer_range"):
            extra_config["initializer_range"] = args.init_std

        if not args.use_mup:
            args.mup_scale_emb = 1.0
            args.mup_scale_depth = math.sqrt(model_config.num_hidden_layers)
            args.mup_scale_width = 1.0
        else:
            if hasattr(model_config, "mup_scale_emb"):
                args.mup_scale_emb = args.mup_scale_emb or model_config.mup_scale_emb
            if hasattr(model_config, "mup_scale_depth"):
                args.mup_scale_depth = args.mup_scale_depth or model_config.mup_scale_depth
            if hasattr(model_config, "mup_scale_width"):
                args.mup_scale_width = args.mup_scale_width or model_config.mup_scale_width

        extra_config["mup_scale_emb"] = args.mup_scale_emb
        extra_config["mup_scale_depth"] = args.mup_scale_depth
        extra_config["mup_scale_width"] = args.mup_scale_width

        if isinstance(extra_config, dict):
            for k, v in extra_config.items():
                if hasattr(model_config, k):
                    model_config.__setattr__(k, v)

        logger.info(f"rank-{rank} -> [start]init model({args.model_clz}) with {model_config}")
        model_clz = importer.from_name_to_clz(args.model_clz)
        if hasattr(model_clz, "from_config"):
            model = model_clz.from_config(
                model_config, use_flash_attention_2=args.use_flash_attn, trust_remote_code=True
            )
        else:
            model = model_clz(model_config)

    if args.grad_checkpointing:
        model.gradient_checkpointing_enable(gradient_checkpointing_kwargs={"use_reentrant": args.use_reentrant})

    architecture = model.__class__.__name__
    model_numel, non_embed_numel, trainable_numel = get_model_numel(model)
    if coordinator.is_master():
        args_dict = {key: value for key, value in args.__dict__.items() if key != "dataset"}
        logger.info(
            f"{extra_info} model specs architecture: {architecture}"
            f", parameters full: {format_numel(model_numel)}"
            f", non-embed: {format_numel(non_embed_numel)}"
            f", trainable: {format_numel(trainable_numel)}"
            f", args: {json.dumps(args_dict, ensure_ascii=False)}"
        )

    optimizer = get_optimizer(args, model) if with_optimizer else None
    lr_scheduler = get_lr_scheduler(args, optimizer) if with_scheduler else None
    booster = get_booster(args)

    model, optimizer, _, _, lr_scheduler = booster.boost(model, optimizer, lr_scheduler=lr_scheduler)

    st_step, extra_states = 1, {}

    if args.load:
        states = storage.load(
            booster,
            model,
            optimizer,
            lr_scheduler,
            args.load,
            coordinator.rank,
            args.reset_states,
            not args.not_use_strict,
        )
        st_step, extra_states = states

    # fmt: off
    return (
        booster, model, optimizer, lr_scheduler,
        architecture, model_config, model_numel,
        st_step, extra_states
    )
    # fmt: on
